import json
import sys

import aiohttp
from cachetools import TTLCache, cached
from fastapi import APIRouter, Depends, HTTPException, Request
from fastapi.responses import HTMLResponse, StreamingResponse

from .. import config, errors, utils
from ..client import AiohttpSingleton

simple_router = APIRouter(
    prefix="/simple",
    tags=["simple"],
    default_response_class=HTMLResponse,
)


@cached(cache=TTLCache(maxsize=sys.maxsize, ttl=30))
def get_special_cases() -> dict[str, str]:
    with open(config.SPECIAL_CASES_FILE_LOCATION, "r") as file:
        return json.load(file)


@simple_router.get("/")
async def simple_package_index(
        request: Request,
        special_cases: dict[str, str] = Depends(get_special_cases),
        session: aiohttp.ClientSession = Depends(AiohttpSingleton.session),
):
    ttl_cache: TTLCache = request.app.state.cache
    try:
        cached_packages = ttl_cache["packages"]
    except KeyError:
        # Get filled when consuming the grouped_packages_generator
        packages: dict[str, str] = {}
        try:
            grouped_packages_generator = utils.generate_grouped_packages(
                package_cache=packages,
                sources=config.SOURCES,
                special_cases=special_cases,
                session=session,
            )
        except errors.MissingHrefError:
            raise HTTPException(
                500,
                "One of the sources contains an entry without a href attribute",
            )
        ttl_cache["packages"] = packages
    else:
        grouped_packages_generator = utils.iterable_to_async_generator(
            cached_packages.items(),
        )

    index_generator = utils.generate_index_from_packages(grouped_packages_generator)

    return StreamingResponse(index_generator)


@simple_router.get("/{package_name}/")
async def simple_package_releases(
        package_name: str,
        special_cases: dict[str, str] = Depends(get_special_cases),
        session=Depends(AiohttpSingleton.session),
):
    if package_name != utils.normalize(package_name):
        raise HTTPException(
            400,
            f"Given package name '{package_name}' is not normalized",
        )
    try:
        return await utils.get_package_releases(
            sources=config.SOURCES,
            package_name=package_name,
            special_cases=special_cases,
            session=session,
        )
    except errors.PackageNotFoundError as e:
        raise HTTPException(404, e.msg)
    except errors.MissingHrefError:
        raise HTTPException(
            500,
            "One of the sources contains an entry without a href attribute",
        )
