from unittest import mock

from fastapi.testclient import TestClient
import pytest

from acc_py_index import config, errors, utils
from acc_py_index.main import app
from acc_py_index.routers.simple import get_special_cases


@pytest.fixture
def client() -> TestClient:
    try:
        app.dependency_overrides[get_special_cases] = lambda: {}
        return TestClient(app)
    finally:
        app.state.cache.pop("packages", default=None)
        app.state.cache.expire()


def test_simple_package_index(monkeypatch, client: TestClient):
    sources = [f"http://source{n}/simple/" for n in range(3)]

    def mocked_get_url_stream(url: str, _session: None):
        if url == sources[0]:
            return utils.iterable_to_async_generator([
                b"<html>\n  <a href=\"banana\">Banana</a>\n  <a ",
                b"href=\"/blue-kiwi\">blue.KIWI</a>\n</html>",
            ])
        elif url == sources[1]:
            return utils.iterable_to_async_generator([
                b"<html>\n  <a href=\"../basket/red-lemon\">red_lemon<",
                b"/a>\n  <a href=\"https://website.org/banana\">ban",
                b"ana</a>\n</html>",
            ])
        elif url == sources[2]:
            return utils.iterable_to_async_generator([
                b"<html>\n<a href=\"https://website.org/blue-k",
                b"iwi\">blue-kiwi</a>\n  <a href=\"https://web",
                b"site.org/red-lemon\">red_lemon</a>\n  <a href",
                b"=\"https://website.org/grape\">grape</a>\n</html>",
            ])
        raise AssertionError("Test URL mismatch")

    monkeypatch.setattr(utils, "get_url_stream", mocked_get_url_stream)
    monkeypatch.setattr(config, "SOURCES", sources)

    response = client.get("/simple/")
    assert response.status_code == 200
    assert response.text == """<!DOCTYPE html>
<html>
  <head>
    <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
    <title>Simple index</title>
  </head>
  <body>
    <a href="http://source0/simple/banana">banana</a><br/>
    <a href="http://source0/blue-kiwi">blue-kiwi</a><br/>
    <a href="http://source1/basket/red-lemon">red-lemon</a><br/>
    <a href="https://website.org/grape">grape</a><br/>
  </body>
</html>"""
    client.app.state.cache.pop("packages")


def test_simple_index_ttl_cache(monkeypatch, client: TestClient):
    packages = {
        "numpy": "http://somewhere/numpy",
        "acc-py": "http://somewhereelse/acc-py",
    }
    monkeypatch.setattr(
        utils,
        "generate_grouped_packages",
        lambda **_kwargs: utils.iterable_to_async_generator(packages.items()),
    )
    assert "packages" not in client.app.state.cache
    client.get("/simple/")
    assert "packages" in client.app.state.cache

    client.app.state.cache["packages"].update(packages)
    response = client.get("/simple/")
    assert response.text == """<!DOCTYPE html>
<html>
  <head>
    <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
    <title>Simple index</title>
  </head>
  <body>
    <a href="http://somewhere/numpy">numpy</a><br/>
    <a href="http://somewhereelse/acc-py">acc-py</a><br/>
  </body>
</html>"""


def test_simple_index_missing_href(monkeypatch, client: TestClient):
    with mock.patch(
            "acc_py_index.utils.generate_grouped_packages",
            side_effect=errors.MissingHrefError,
    ):
        response = client.get("/simple/")

    assert response.status_code == 500
    assert response.json() == {
        "detail": "One of the sources contains an entry without a href attribute",
    }


def test_simple_package_releases_not_normalized(monkeypatch, client: TestClient):
    response = client.get("/simple/not_normalized")
    assert response.status_code == 400
    assert response.json() == {
        "detail": "Given package name 'not_normalized' is not normalized",
    }


def test_simple_package_releases_package_not_found(monkeypatch, client: TestClient):
    with mock.patch(
            "acc_py_index.utils.get_package_releases",
            side_effect=errors.PackageNotFoundError(
                package_name="ghost", sources=["source1", "source2"],
            ),
    ):
        response = client.get("/simple/ghost")
    assert response.status_code == 404
    assert response.json() == {
        'detail': "Package 'ghost' was not found in "
                  "the configured sources: ['source1', 'source2']",
    }


def test_simple_package_releases_missing_href(monkeypatch, client: TestClient):
    with mock.patch(
            "acc_py_index.utils.get_package_releases",
            side_effect=errors.MissingHrefError,
    ):
        response = client.get("/simple/any")

    assert response.status_code == 500
    assert response.json() == {
        'detail': "One of the sources contains an entry without a href attribute",
    }
