from unittest import mock

import pytest

from acc_py_index import errors, utils
from acc_py_index.tests.aiohttp_mock import MockedRequestContextManager


@pytest.mark.asyncio
async def test_iterable_to_async_generator():
    assert [0, 1, 2] == [i async for i in utils.iterable_to_async_generator(range(3))]


def test_normalize():
    assert utils.normalize("A.b-C_D") == "a-b-c-d"


@pytest.mark.asyncio
async def test_get_url_stream(monkeypatch):
    chunks = [
        b"<html>\n<bo",
        b"dy>\n  <",
        b'a href=\"/test',
        b"\">TEST</a>\n</body",
        b"></html>",
    ]

    response_mock = mock.MagicMock()
    response_mock.content.iter_chunks.return_value = utils.iterable_to_async_generator(
        (chunk, False) for chunk in chunks
    )
    session_mock = mock.Mock()
    session_mock.get.return_value = MockedRequestContextManager(response_mock)
    url_stream = utils.get_url_stream("", session_mock)

    assert chunks == [chunk async for chunk in url_stream]


@pytest.mark.asyncio
async def test_extract_packages_from_html_stream():
    chunks = [
        b"<html>\n  <a href=\"http://te",
        b"st/simple/pytest-aiohttp\">PytEst_aioH",
        b"TTP</a>\n  <a href=\"http://test/fastapi-utils\">fast",
        b"api.utils</a>\n  <a hre",
        b"f=\"http://test/packages/pyjapc\">PYJAPC</",
        b"a>\n  <a href=\"https://website.org/sqlalchemy-pydantic-orm\">",
        b"sqlalchemy-pydantic-orm</a>\n</html>",
    ]
    stream = utils.iterable_to_async_generator(chunks)
    packages = {
        package: href
        async for package, href in utils.extract_packages_from_html_stream(stream)
    }
    assert packages == {
        'pytest-aiohttp': 'http://test/simple/pytest-aiohttp',
        'fastapi-utils': 'http://test/fastapi-utils',
        'pyjapc': 'http://test/packages/pyjapc',
        'sqlalchemy-pydantic-orm': 'https://website.org/sqlalchemy-pydantic-orm',
    }


@pytest.mark.asyncio
async def test_extract_packages_from_html_stream_missing_href_error():
    chunks = [
        b"<html><a hr",
        b"ef=\"numpy\">numpy</a>\n  <a>pyt",
        b"imber</a></html>",
    ]
    html_stream = utils.iterable_to_async_generator(chunks)
    package_stream = utils.extract_packages_from_html_stream(html_stream)
    await package_stream.__anext__()
    with pytest.raises(errors.MissingHrefError):
        await package_stream.__anext__()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("special_cases", "expected"), [
        (
                {},
                {
                    "blue-kiwi": "http://source0/blue-kiwi",
                    "banana": "http://source0/simple/banana",
                    "red-lemon": "http://source1/basket/red-lemon",
                    "grape": "https://website.org/grape",
                },
        ),
        (
                {"banana": "http://special/location/banana"},
                {
                    "blue-kiwi": "http://source0/blue-kiwi",
                    "banana": "http://special/location/banana",
                    "red-lemon": "http://source1/basket/red-lemon",
                    "grape": "https://website.org/grape",
                },
        ),
        (
                {"grape": "http://special/location/grape"},
                {
                    "blue-kiwi": "http://source0/blue-kiwi",
                    "banana": "http://source0/simple/banana",
                    "red-lemon": "http://source1/basket/red-lemon",
                    "grape": "http://special/location/grape",
                },
        ),
    ],
)
async def test_generate_grouped_packages(monkeypatch, special_cases, expected):
    sources = [f"http://source{n}/simple/" for n in range(3)]

    def mocked_get_url_stream(url: str, _session: None):
        if url == sources[0]:
            return utils.iterable_to_async_generator([
                b"<html>\n  <a href=\"/blue-kiwi\">blue.KI",
                b"WI</a>\n  <a href=\"banana\">Banana</a>\n</html>",
            ])
        elif url == sources[1]:
            return utils.iterable_to_async_generator([
                b"<html><a href=\"../basket/re",
                b"d-lemon\">red_lemon</a><a hre",
                b"f=\"https://website.org/banana\">banana</a></html>",
            ])
        elif url == sources[2]:
            return utils.iterable_to_async_generator([
                b"<html>\n  <a href=\"https://website.org/red-lemon",
                b"\">red_lemon</a>\n  <a href=\"https://website.org/grape\">grape</a>",
                b"\n  <a href=\"https://website.org/blue-kiwi\">blue-kiwi</a>\n</html>",
            ])
        raise AssertionError("Test URL mismatch")

    monkeypatch.setattr(utils, "get_url_stream", mocked_get_url_stream)

    package_cache = {}
    grouped_packages_stream = utils.generate_grouped_packages(
        package_cache=package_cache,
        sources=sources,
        special_cases=special_cases,
        session=...,
    )
    grouped_packages = {
        package: href
        async for package, href in grouped_packages_stream
    }
    assert grouped_packages == expected
    assert package_cache == expected


@pytest.mark.asyncio
async def test_generate_index_from_packages():
    packages = (
        ("blue-kiwi", "http://source0/blue-kiwi"),
        ("banana", "http://source0/simple/banana"),
        ("red-lemon", "http://source1/basket/red-lemon"),
        ("grape", "https://website.org/grape"),
    )
    package_stream = utils.iterable_to_async_generator(packages)
    index_generator = utils.generate_index_from_packages(package_stream)
    index_html = "".join([row async for row in index_generator])
    assert index_html == """<!DOCTYPE html>
<html>
  <head>
    <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
    <title>Simple index</title>
  </head>
  <body>
    <a href="http://source0/blue-kiwi">blue-kiwi</a><br/>
    <a href="http://source0/simple/banana">banana</a><br/>
    <a href="http://source1/basket/red-lemon">red-lemon</a><br/>
    <a href="https://website.org/grape">grape</a><br/>
  </body>
</html>"""
