import asyncio
from unittest import mock

import pytest

from acc_py_index import errors, utils
from acc_py_index.tests.aiohttp_mock import MockedRequestContextManager


def test_html_href_absolutizer():
    html_input = """
<html>
<a href="numpy"/>
<a href="/pytimber"/>
<a href="../packages/pyjapc"/>
<a href="https://website.org/pyrbac"/>
</html>
"""
    html_output = utils.html_href_absolutizer(html_input, "http://test/simple/")
    assert html_output == """<html>
<body><a href="http://test/simple/numpy"></a>
<a href="http://test/pytimber"></a>
<a href="http://test/packages/pyjapc"></a>
<a href="https://website.org/pyrbac"></a>
</body></html>
"""


def test_html_href_absolutizer_missing_href_error():
    missing_href_html = """
<html>
<a href="numpy"/>
<a>pytimber</a>
</html>
"""
    with pytest.raises(errors.MissingHrefError):
        utils.html_href_absolutizer(missing_href_html, "http://test/simple/")


@pytest.mark.asyncio
async def test_get_package_releases(monkeypatch):
    sources = [f"http://source{n}/simple/" for n in range(2)]
    mocked_session = mock.MagicMock()

    def mocked_session_get(url: str, **_kwargs):
        assert url == str(sources[0] + "pytimber/")
        response_mock = mock.MagicMock()
        response_mock.status = 200
        future: asyncio.Future = asyncio.Future()
        future.set_result("""<html>
<a href="banana">Banana</a>
<a href="/blue-kiwi">blue.KIWI</a>
</html>""")
        response_mock.text.return_value = future
        return MockedRequestContextManager(response_mock)

    mocked_session.get = mocked_session_get
    package_releases = await utils.get_package_releases(
        sources=sources,
        package_name="pytimber",
        special_cases={},
        session=mocked_session,
    )
    assert package_releases == """<!DOCTYPE html>
<html>
<body><a href="http://source0/simple/pytimber/banana">Banana</a>
<a href="http://source0/blue-kiwi">blue.KIWI</a>
</body></html>"""


@pytest.mark.asyncio
async def test_get_package_releases_package_not_found_error(monkeypatch):
    sources = [f"http://source{n}/simple/" for n in range(2)]
    response_mock = mock.MagicMock()
    response_mock.status = 404
    mocked_session = mock.MagicMock()
    mocked_session.get.return_value = MockedRequestContextManager(response_mock)

    with pytest.raises(
            expected_exception=errors.PackageNotFoundError,
            match=r"Package 'ghost' was not found in the configured sources: "
                  r"\['http://source0/simple/', 'http://source1/simple/'\]",
    ):
        await utils.get_package_releases(
            sources=sources,
            package_name="ghost",
            special_cases={},
            session=mocked_session,
        )
