import typing
from urllib.parse import urljoin

import aiohttp
from bs4 import BeautifulSoup
from lxml import etree

from . import errors

T = typing.TypeVar("T")

SIMPLE_INDEX_HEADER = """<!DOCTYPE html>
<html>
  <head>
    <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
    <title>Simple index</title>
  </head>
  <body>
"""
SIMPLE_INDEX_PACKAGE_LINK = '    <a href="{href}">{package}</a><br/>\n'
SIMPLE_INDEX_FOOTER = "  </body>\n</html>"


async def iterable_to_async_generator(
        iterable: typing.Iterable[T],
) -> typing.AsyncGenerator[T, None]:
    for chunk in iterable:
        yield chunk


def normalize(package_name: str) -> str:
    """Normalizes package name according to PEP 503 (without regex)"""
    return package_name.lower().replace(".", "-").replace("_", "-")


async def get_url_stream(
        url: str,
        session: aiohttp.ClientSession,
) -> typing.AsyncGenerator[bytes, None]:
    """Retrieves a stream of chunks from a given URL"""
    async with session.get(url) as resp:
        # `resp.content.iter_chunked()` can be used to manually supply a max chunk_size.
        # Remove index of 0 on `chunk` when using `.iter_chunked()`
        async for chunk in resp.content.iter_chunks():
            yield chunk[0]


async def extract_packages_from_html_stream(
        stream: typing.AsyncGenerator[bytes, None],
) -> typing.AsyncGenerator[tuple[str, str], None]:
    """Incrementally parses HTML chunks to yield package names and links"""
    parser = etree.HTMLPullParser(tag="a")
    async for chunk in stream:
        parser.feed(chunk)
        for _event, element in parser.read_events():
            try:
                href = element.attrib["href"]
            except KeyError:
                raise errors.MissingHrefError
            yield normalize(element.text), href


async def generate_grouped_packages(
        package_cache: dict[str, str],
        sources: list[str],
        special_cases: dict[str, str],
        session: aiohttp.ClientSession,
) -> typing.AsyncGenerator[tuple[str, str], None]:
    """Generates HTML with the merged indices of the given sources"""
    for source in sources:
        source_stream = get_url_stream(source, session)
        async for package, href in extract_packages_from_html_stream(source_stream):
            if package not in package_cache:
                try:
                    absolute_href = special_cases[package]
                except KeyError:
                    absolute_href = urljoin(source, href)
                package_cache[package] = absolute_href
                yield package, absolute_href


async def generate_index_from_packages(
        packages: typing.AsyncGenerator[tuple[str, str], None],
) -> typing.AsyncGenerator[str, None]:
    """Generates index from packages saved in cache"""
    yield SIMPLE_INDEX_HEADER
    async for package, href in packages:
        yield SIMPLE_INDEX_PACKAGE_LINK.format(package=package, href=href)
    yield SIMPLE_INDEX_FOOTER


def html_href_absolutizer(html: str, url_base: str) -> str:
    """Converts all relative hrefs in a given html to absolute urls"""
    soup = BeautifulSoup(html, "lxml")
    for link in soup.find_all("a"):
        try:
            if not link["href"].startswith("http"):
                link["href"] = urljoin(url_base, link["href"])
        except KeyError:
            raise errors.MissingHrefError
    return str(soup)


async def get_package_releases(
        sources: list[str],
        package_name: str,
        special_cases: dict[str, str],
        session: aiohttp.ClientSession,
) -> typing.Optional[str]:
    """Gets package releases from the given sources in order"""
    for source in sources:
        try:
            url = special_cases[package_name]
        except KeyError:
            url = source + package_name + "/"
        async with session.get(url) as resp:
            if resp.status < 400:
                body: str = await resp.text()
                if not body.lower().lstrip().startswith("<!DOCTYPE html>"):
                    # Temporary fix: https://github.com/pypa/pip/issues/10825
                    body = "<!DOCTYPE html>\n" + body
                return html_href_absolutizer(body, url)

    raise errors.PackageNotFoundError(package_name, sources)
